@@ -36,6 +36,25 @@ EMAIL_MAX_LENGTH = 254
 LOCAL_PART_MAX_LENGTH = 64
 DOMAIN_MAX_LENGTH = 255
 
+# IANA Special Use Domain Names
+# Last Updated 2021-09-21
+# https://www.iana.org/assignments/special-use-domain-names/special-use-domain-names.txt
+# The domain names without dots would be caught by the check that the domain
+# name in an email address must have a period, but this list will also catch
+# subdomains of these domains, which are also reserved.
+SPECIAL_USE_DOMAIN_NAMES = (
+    "arpa",  # consolidated from a lot of arpa subdomains, we'll assume all subdomains of arpa are actually reserved
+    "example",
+    "example.com",
+    "example.net",
+    "example.org",
+    "invalid",
+    "local",
+    "localhost",
+    "onion",
+    "test",  # see special logic for 'test' where this is checked
+)
+
 # ease compatibility in type checking
 if sys.version_info >= (3,):
     unicode_class = str
@@ -142,6 +161,8 @@ class ValidatedEmail(object):
 
     """Tests use this."""
     def __eq__(self, other):
+        if not isinstance(other, ValidatedEmail):
+            return False
         return (
             self.email == other.email
             and self.local_part == other.local_part
@@ -192,6 +213,7 @@ def validate_email(
     allow_smtputf8=True,
     allow_empty_local=False,
     check_deliverability=True,
+    test_environment=False,
     timeout=DEFAULT_TIMEOUT,
     dns_resolver=None
 ):
@@ -228,7 +250,7 @@ def validate_email(
     ret.smtputf8 = local_part_info["smtputf8"]
 
     # Validate the email address's domain part syntax and get a normalized form.
-    domain_part_info = validate_email_domain_part(parts[1])
+    domain_part_info = validate_email_domain_part(parts[1], test_environment=test_environment)
     ret.domain = domain_part_info["domain"]
     ret.ascii_domain = domain_part_info["ascii_domain"]
 
@@ -278,9 +300,9 @@ def validate_email(
             reason = "(when encoded in bytes)"
         raise EmailSyntaxError("The email address is too long {}.".format(reason))
 
-    if check_deliverability:
-        # Validate the email address's deliverability and update the
-        # return dict with metadata.
+    if check_deliverability and not test_environment:
+        # Validate the email address's deliverability using DNS
+        # and update the return dict with metadata.
         deliverability_info = validate_email_deliverability(
             ret["domain"], ret["domain_i18n"], timeout, dns_resolver
         )
@@ -332,7 +354,7 @@ def validate_email_local_part(local, allow_smtputf8=True, allow_empty_local=Fals
         if not m:
             # It's not a valid internationalized address either. Report which characters were not valid.
             bad_chars = ', '.join(sorted(set(
-                c for c in local if not re.match(u"[" + (ATEXT if not allow_smtputf8 else ATEXT_UTF8) + u"]", c)
+                unicodedata.name(c, repr(c)) for c in local if not re.match(u"[" + (ATEXT if not allow_smtputf8 else ATEXT_UTF8) + u"]", c)
             )))
             raise EmailSyntaxError("The email address contains invalid characters before the @-sign: %s." % bad_chars)
 
@@ -346,6 +368,40 @@ def validate_email_local_part(local, allow_smtputf8=True, allow_empty_local=Fals
         # so we'll return the normalized local part in the return value.
         local = unicodedata.normalize("NFC", local)
 
+        # Check for unsafe characters.
+        # Some of this may be redundant with the range U+0080 to U+10FFFF that is checked
+        # by DOT_ATOM_TEXT_UTF8.
+        for i, c in enumerate(local):
+            category = unicodedata.category(c)
+            if category[0] in ("L", "N", "P", "S"):
+                # letters, numbers, punctuation, and symbols are permitted
+                pass
+            elif category[0] == "M":
+                # combining character in first position would combine with something
+                # outside of the email address if concatenated to the right, but are
+                # otherwise permitted
+                if i == 0:
+                    raise EmailSyntaxError("The email address contains an initial invalid character (%s)."
+                                           % unicodedata.name(c, repr(c)))
+            elif category[0] in ("Z", "C"):
+                # spaces and line/paragraph characters (Z) and
+                # control, format, surrogate, private use, and unassigned code points (C)
+                raise EmailSyntaxError("The email address contains an invalid character (%s)."
+                                       % unicodedata.name(c, repr(c)))
+            else:
+                # All categories should be handled above, but in case there is something new
+                # in the future.
+                raise EmailSyntaxError("The email address contains a character (%s; category %s) that may not be safe."
+                                       % (unicodedata.name(c, repr(c)), category))
+
+        # Try encoding to UTF-8. Failure is possible with some characters like
+        # surrogate code points, but those are checked above. Still, we don't
+        # want to have an unhandled exception later.
+        try:
+            local.encode("utf8")
+        except ValueError:
+            raise EmailSyntaxError("The email address contains an invalid character.")
+
         # Flag that SMTPUTF8 will be required for deliverability.
         return {
             "local_part": local,
@@ -354,7 +410,7 @@ def validate_email_local_part(local, allow_smtputf8=True, allow_empty_local=Fals
         }
 
 
-def validate_email_domain_part(domain):
+def validate_email_domain_part(domain, test_environment=False):
     # Empty?
     if len(domain) == 0:
         raise EmailSyntaxError("There must be something after the @-sign.")
@@ -433,11 +489,33 @@ def validate_email_domain_part(domain):
         raise EmailSyntaxError("The email address contains invalid characters after the @-sign.")
 
     # All publicly deliverable addresses have domain named with at least
-    # one period. We also know that all TLDs end with a letter.
-    if "." not in ascii_domain:
+    # one period, and we'll consider the lack of a period a syntax error
+    # since that will match people's sense of what an email address looks
+    # like. We'll skip this in test environments to allow '@test' email
+    # addresses.
+    if "." not in ascii_domain and not (ascii_domain == "test" and test_environment):
         raise EmailSyntaxError("The domain name %s is not valid. It should have a period." % domain_i18n)
+
+    # Check special-use and reserved domain names. Raise these as
+    # deliverability errors since they are syntactically valid.
+    # Some might fail DNS-based deliverability checks, but that
+    # can be turned off, so we should fail them all sooner.
+    for d in SPECIAL_USE_DOMAIN_NAMES:
+        # RFC 6761 says that applications should not block use of the 'test'
+        # domain name, presumably because that would prevent it from being
+        # used for actual testing. We'll block it, except when a special
+        # testing flag is used, indicating that the module is being used
+        # in a test environment.
+        if d == "test" and test_environment:
+            continue
+
+        if ascii_domain == d or ascii_domain.endswith("." + d):
+            raise EmailUndeliverableError("The domain name %s is a special-use or reserved name that cannot be used with email." % domain_i18n)
+
+    # We also know that all TLDs currently end with a letter, and
+    # we'll consider that a non-DNS based deliverability check.
     if not re.search(r"[A-Za-z]\Z", ascii_domain):
-        raise EmailSyntaxError(
+        raise EmailUndeliverableError(
             "The domain name %s is not valid. It is not within a valid top-level domain." % domain_i18n
         )
 
@@ -487,10 +565,19 @@ def validate_email_deliverability(domain, domain_i18n, timeout=DEFAULT_TIMEOUT,
             raise dns.exception.Timeout()
 
         try:
-            # Try resolving for MX records and get them in sorted priority order.
+            # Try resolving for MX records and get them in sorted priority order
+            # as (priority, qname) pairs.
             response = dns_resolver_resolve_shim(domain, "MX")
             mtas = sorted([(r.preference, str(r.exchange).rstrip('.')) for r in response])
             mx_fallback = None
+
+            # Do not permit delivery if there is only a "null MX" record (whose value is
+            # (0, ".") but we've stripped trailing dots, so the 'exchange' is just "").
+            mtas = [(preference, exchange) for preference, exchange in mtas
+                    if exchange != ""]
+            if len(mtas) == 0:
+                raise EmailUndeliverableError("The domain name %s does not accept email." % domain_i18n)
+
         except (dns.resolver.NoNameservers, dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
 
             # If there was no MX record, fall back to an A record.
@@ -534,7 +621,6 @@ def validate_email_deliverability(domain, domain_i18n, timeout=DEFAULT_TIMEOUT,
 
 
 def main():
-    import sys
     import json
 
     def __utf8_input_shim(input_str):
@@ -548,10 +634,12 @@ def main():
         return output_str
 
     if len(sys.argv) == 1:
+        # Validate the email addresses pased line-by-line on STDIN.
+        dns_resolver = caching_resolver()
         for line in sys.stdin:
             email = __utf8_input_shim(line.strip())
             try:
-                validate_email(email)
+                validate_email(email, dns_resolver=dns_resolver)
             except EmailNotValidError as e:
                 print(__utf8_output_shim("{} {}".format(email, e)))
     else:
